"""Terminal helper"""

__docformat__ = "numpy"

import hashlib
import logging
import os
import sys
import webbrowser

# IMPORTATION STANDARD
from contextlib import contextmanager
from typing import Any, Dict, List, Optional

import matplotlib.pyplot as plt

# IMPORTATION THIRDPARTY
from packaging import version

# IMPORTATION INTERNAL
import infomark_terminal.core.session.local_model as Local
from infomark_terminal import thought_of_the_day as thought
from infomark_terminal.base_helpers import load_env_files
from infomark_terminal.core.config.paths import HIST_FILE_PATH, SETTINGS_ENV_FILE
from infomark_terminal.core.plots.backend import plots_backend
from infomark_terminal.core.session.constants import BackendEnvironment
from infomark_terminal.core.session.current_system import get_current_system
from infomark_terminal.core.session.current_user import (
    get_current_user,
    is_local,
    set_preference,
)
from infomark_terminal.core.session.env_handler import write_to_dotenv
from infomark_terminal.helper_funcs import request
from infomark_terminal.rich_config import console

# pylint: disable=too-many-statements,no-member,too-many-branches,C0302

try:
    __import__("git")
except ImportError:
    WITH_GIT = False
else:
    WITH_GIT = True
logger = logging.getLogger(__name__)


def print_goodbye():
    """Prints a goodbye message when quitting the terminal"""
    # LEGACY GOODBYE MESSAGES - You'll live in our hearts forever.
    # "An informed ape, is a strong ape."
    # "Remember that stonks only go up."
    # "Diamond hands."
    # "Apes together strong."
    # "This is our way."
    # "Keep the spacesuit ape, we haven't reached the moon yet."
    # "I am not a cat. I'm an ape."
    # "We like the terminal."
    # "...when offered a flight to the moon, nobody asks about what seat."

    text = """
[param]Thank you for using the Infomark Terminal and being part of this journey.[/param]

We hope you'll find the new CLI as valuable as this. To stay tuned, sign up for our newsletter: [cmds]https://Infomark.co/newsletter.[/]

In the meantime, check out our other products:

[bold]Infomark Terminal Pro[/]: [cmds]https://Infomark.co/products/pro[/cmds]
[bold]Infomark Platform:[/]     [cmds]https://Infomark.co/products/platform[/cmds]
[bold]Infomark Bot[/]:          [cmds]https://Infomark.co/products/bot[/cmds]
    """
    console.print(text)
    logger.info("END")


def sha256sum(filename):
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    return h.hexdigest()


def open_Infomark_documentation(  # noqa: PLR0912
    path,
    url="https://docs.Infomark.co/terminal",
    command=None,
    arg_type="",
):
    """Opens the documentation page based on your current location within the terminal. Make exceptions for menus
    that are considered 'common' by adjusting the path accordingly."""
    if path == "/" and command is None:
        path = "/usage/overview/structure-and-navigation"
        command = ""
    elif "keys" in path:
        path = "/usage/data/api-keys"
        command = ""
    elif "settings" in path:
        path = "/usage/overview/customizing-the-terminal"
        command = ""
    elif "featflags" in path:
        path = "/usage/overview/customizing-the-terminal#feature-flags-menu"
        command = ""
    elif "sources" in path:
        path = "/usage/usage/data/data-sources"
        command = ""
    elif "account" in path:
        path = "/usage/hub"
        command = ""
    elif arg_type == "command":  # user passed a command name
        if command in ["settings", "featflags"]:
            path = "/usage/overview/customizing-the-terminal"
            command = ""
        else:
            path = f"/reference/{path}"
    elif "askobb" in path:
        path = "/usage/askobb-feature"
        command = ""
    elif arg_type == "menu":  # user passed a menu name
        if command in ["ta", "qa"]:
            menu = path.split("/")[-2]
            path = f"/menus/common/{menu}"
        elif command == "stocks":
            path = "/menus/stocks/introduction"
            command = ""
        elif command == "forecast":
            command = ""
            path = "/menus/forecast"
        elif command == "crypto":
            path = "/menus/crypto/introduction"
            command = ""

        else:
            path = f"/menus/{path}"
    else:  # user didn't pass argument and is in a menu
        menu = path.split("/")[-2]
        if menu == "crypto" and not command:
            path = "/crypto/introduction"
        if menu == "stocks":
            path = "/stocks/introduction"
        path = f"/menus/common/{menu}" if menu in ["ta", "qa"] else f"/menus/{path}"

    if command:
        if command == "keys":
            path = "/usage/data/api-keys"
            command = ""
        elif "settings" in path or "featflags" in path:
            path = "/usage/overview/customizing-the-terminal"
            command = ""
        elif "sources" in path:
            path = "/usage/data/data-sources"
            command = ""
        elif command in ["record", "stop", "exe"]:
            path = "/usage/routines/introduction-to-routines"
            command = ""
        elif command == "sources":
            path = "/usage/data/data-sources"
            command = ""
        elif command == "askobb":
            path = "/usage/askobb-feature"
            command = ""
        elif command == "news":
            path = "/usage/overview/commands-and-arguments#help-arguments"
            command = ""
        elif command in [
            "intro",
            "about",
            "support",
            "survey",
            "update",
            "wiki",
        ]:
            path = "/usage"
            command = ""
        elif command in ["ta", "qa"]:
            path = f"/menus/common/{command}"
            command = ""
        elif command == "stocks":
            path = "/menus/stocks/introduction"
            command = ""
        elif command == "account":
            path = "/usage/hub"
            command = ""
        elif command == "news":
            path = "/usage/overview/commands-and-arguments#help-arguments"
            command = ""
        path += command

    full_url = f"{url}{path.replace('//', '/')}"

    if full_url[-1] == "/":
        full_url = full_url[:-1]

    webbrowser.open(full_url)


def hide_splashscreen():
    """Hide the splashscreen on Windows bundles.

    `pyi_splash` is a PyInstaller "fake-package" that's used to communicate
    with the splashscreen on Windows.
    Sending the `close` signal to the splash screen is required.
    The splash screen remains open until this function is called or the Python
    program is terminated.
    """
    try:
        import pyi_splash  # type: ignore  # pylint: disable=import-outside-toplevel

        pyi_splash.update_text("Terminal Loaded!")
        pyi_splash.close()
    except Exception as e:
        logger.info(e)


def is_auth_enabled() -> bool:
    """Tell whether or not authentication is enabled.

    Returns
    -------
    bool
        If authentication is enabled
    """
    # TODO: This function is a temporary way to block authentication
    return get_current_system().ENABLE_AUTHENTICATION


def print_guest_block_msg():
    """Block guest users from using the terminal."""
    if is_local():
        console.print(
            "[info]You are currently logged as a guest.[/info]\n"
            "[info]Login to use this feature.[/info]\n\n"
            "[info]If you don't have an account, you can create one here: [/info]"
            f"[cmds]{BackendEnvironment.HUB_URL + 'register'}\n[/cmds]"
        )


def is_installer() -> bool:
    """Tell whether or not it is a packaged version (Windows or Mac installer"""
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def bootup():
    if sys.platform == "win32":
        # Enable VT100 Escape Sequence for WINDOWS 10 Ver. 1607
        os.system("")  # nosec # noqa: S605,S607
        # Hide splashscreen loader of the packaged app
        if is_installer():
            hide_splashscreen()

    try:
        if os.name == "nt":
            # pylint: disable=E1101
            sys.stdin.reconfigure(encoding="utf-8")
            # pylint: disable=E1101
            sys.stdout.reconfigure(encoding="utf-8")
    except Exception as e:
        logger.exception("Exception: %s", str(e))
        console.print(e, "\n")


def check_for_updates() -> None:
    """Check if the latest version is running.

    Checks github for the latest release version and compares it to cfg.VERSION.
    """
    # The commit has was commented out because the terminal was crashing due to git import for multiple users
    # ({str(git.Repo('.').head.commit)[:7]})
    try:
        r = request(
            "https://api.github.com/repos/Infomark-finance/Infomarkterminal/releases/latest"
        )
    except Exception:
        r = None

    if r is not None and r.status_code == 200:
        latest_tag_name = r.json()["tag_name"]
        latest_version = version.parse(latest_tag_name)
        current_version = version.parse(get_current_system().VERSION)

        if check_valid_versions(latest_version, current_version):
            if current_version == latest_version:
                console.print("[green]You are using the latest stable version[/green]")
            else:
                console.print(
                    "[yellow]You are not using the latest stable version[/yellow]"
                )
                if current_version < latest_version:
                    console.print(
                        "[yellow]Check for updates at https://my.Infomark.co/app/terminal/download[/yellow]"
                    )

                else:
                    console.print(
                        "[yellow]You are using an unreleased version[/yellow]"
                    )

        else:
            console.print("[red]You are using an unrecognized version.[/red]")
    else:
        console.print(
            "[yellow]Unable to check for updates... "
            + "Check your internet connection and try again...[/yellow]"
        )
    console.print("\n")


def check_valid_versions(
    latest_version: version.Version,
    current_version: version.Version,
) -> bool:
    if (
        not latest_version
        or not current_version
        or not isinstance(latest_version, version.Version)
        or not isinstance(current_version, version.Version)
    ):
        return False
    return True


def welcome_message():
    """Print the welcome message

    Prints first welcome message, help and a notification if updates are available.
    """
    console.print(f"\nWelcome to Infomark Terminal v{get_current_system().VERSION}")

    if get_current_user().preferences.ENABLE_THOUGHTS_DAY:
        console.print("---------------------------------")
        try:
            thought.get_thought_of_the_day()
        except Exception as e:
            logger.exception("Exception: %s", str(e))
            console.print(e)


def reset(queue: Optional[List[str]] = None):
    """Resets the terminal.  Allows for checking code without quitting"""
    console.print("resetting...")
    logger.info("resetting")
    plt.close("all")
    plots_backend().close(reset=True)
    load_env_files()
    debug = get_current_system().DEBUG_MODE
    dev = get_current_system().DEV_BACKEND

    try:
        # save the current user
        user_profile = get_current_user().profile
        session: Dict[str, Any] = {
            "access_token": user_profile.token,
            "token_type": user_profile.token_type,
            "uuid": user_profile.uuid,
            "username": user_profile.username,
            "remember": user_profile.remember,
        }

        # remove the hub routines
        if not is_local():
            Local.remove(get_current_user().preferences.USER_ROUTINES_DIRECTORY / "hub")
            if not get_current_user().profile.remember:
                Local.remove(HIST_FILE_PATH)

        # we clear all infomark_terminal modules from sys.modules
        for module in list(sys.modules.keys()):
            parts = module.split(".")
            if parts[0] == "infomark_terminal":
                del sys.modules[module]

        queue_list = ["/".join(queue) if len(queue) > 0 else ""]  # type: ignore
        # pylint: disable=import-outside-toplevel
        # we run the terminal again
        if is_local():
            from infomark_terminal.terminal_controller import main

            main(debug, dev, queue_list, module="")  # type: ignore
        else:
            from infomark_terminal.core.session import session_controller

            session_controller.main(session, queue=queue_list)

    except Exception as e:
        logger.exception("Exception: %s", str(e))
        console.print("Unfortunately, resetting wasn't possible!\n")
        print_goodbye()


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = devnull
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


def first_time_user() -> bool:
    """Whether a user is a first time user. A first time user is someone with an empty .env file.
    If this is true, it also adds an env variable to make sure this does not run again.

    Returns
    -------
    bool
        Whether or not the user is a first time user
    """
    if SETTINGS_ENV_FILE.stat().st_size == 0:
        set_preference("PREVIOUS_USE", True)
        write_to_dotenv("Infomark_PREVIOUS_USE", "True")
        return True
    return False